/**
 * R/SFC Time Series interface
 * (c) 2008 The Research Kitchen Limited
 * Rory Winston <rory@theresearchkitchen.com>
 */

#include <iostream>
#include <vector>
#include <string>
#include <typeinfo>
#include <map>
#include <sys/types.h>

#include "rtr/selectni.h"        
#include "rtr/tibrtsvc.h"       
#include "rtr/sslrtrs.h"        
#include "rtr/ts1srvc.h"        
#include "rtr/objid.h"          
#include "rtr/cmdline.h"        
#include "tsclient/tsbycount.h"
#include "rtr/tconnect.h"        
#include "rtr/rtxfdb.h"           
#include "rtr/dfltlog.h"        

using namespace std;

#include <R.h>
#include <Rinternals.h>
#include <R_ext/Rdynload.h>

/* Utility function */
RTRTimeSeriesService::Period mnemonicToPeriod(const char* mnemonic);

/* Used to flag whether we need to re-read message dictionary */
static bool initialLoad = TRUE;	
static string configPathStr;
static string serviceNameStr;
static string appIdStr;
static string serverNameStr;

/* Ensure that the SEXP is a valid string type */
static bool checkString(SEXP var) {
	return (isString(var) && TYPEOF(STRING_ELT(var,0))==CHARSXP);
}

extern "C" {
	/**
	* R dyn.load entrypoint 
	* Possible initialization can go here
	*/
	__declspec(dllexport) void R_init_mylib(DllInfo* info) {
	}

	/**
	* Initialize underlying configuration
	*/
	__declspec(dllexport) SEXP init(SEXP serviceName, SEXP configPath, SEXP appId, SEXP serverName) {
		if (checkString(serviceName)) {
			serviceNameStr = CHAR(STRING_ELT(serviceName,0));
			Rprintf("serviceName ok: %s\n", serviceNameStr.c_str());
		}
		else
			Rf_error("serviceName must be a valid string!");

		if (checkString(configPath)) {
			configPathStr = CHAR(STRING_ELT(configPath,0));
			Rprintf("configPath ok: %s\n", configPathStr.c_str());
		}
		else
			Rf_error("configPath must be a valid string!");

		if (checkString(appId)) {
			appIdStr = CHAR(STRING_ELT(appId,0));
			Rprintf("appId ok: %s\n", appIdStr.c_str());
		}
		else
			Rf_error("appId must be a valid string!");

		if (checkString(serverName)) {
			serverNameStr = CHAR(STRING_ELT(serverName,0));
			Rprintf("serverName ok: %s\n", serverNameStr.c_str());
		}
		else
			Rf_error("serverName must be specified");
		
		return ScalarInteger(0);
	}
	

	/*
	* Entrypoint for historical data retrieval.
	* Returns a data frame
	*/
	__declspec(dllexport) SEXP fetch(SEXP item, SEXP nItems, SEXP intervalCode, SEXP startDate, SEXP endDate, SEXP debug) {
		RTRString configPath(configPathStr.c_str());
		RTRString serviceName(serviceNameStr.c_str());
		RTRObjectId appId(appIdStr.c_str());
		
		long num = 0;
		const char* start;
		const char* end;
		bool debugOn = false;

		if (isLogical(debug)) {
			int dbg = *(LOGICAL(debug));
			if (dbg == 1)
				debugOn = true;
		}

		/* Time series granularity parameter */
		/* t, d, c, w, m, q, o */	
		/* SFC doesnt handle intraday (or annual data) */
		const char* mnemonic; 
		if (checkString(intervalCode))
			mnemonic = CHAR(STRING_ELT(intervalCode,0));
		else
			mnemonic = "d";	
		
		RTRTimeSeriesService::Period period = mnemonicToPeriod(mnemonic);

		/* Market data item */
		const char* itemName;
		if (checkString(item))
			itemName = CHAR(STRING_ELT(item,0));
		else
			Rf_error("Item name is not a valid string!");

		if (isNumeric(nItems)) 
			num = (long)*(::REAL(nItems));

		if (num == 0) {
			if (checkString(startDate))
				start = CHAR(STRING_ELT(startDate, 0));
			else
				Rf_error("start date is not a valid mm/dd/yyyy date string!");

			if (checkString(endDate))
				end = CHAR(STRING_ELT(endDate, 0));
			else
				Rf_error("end date is not a valid mm/dd/yyyy date string!");
		}

		RTRString symbol(itemName);
		RTRXFileDb* configDb = new RTRXFileDb(configPath);
		if (configDb->error()) {
			Rf_error("Error initializing SFC: %s", configDb->errorText());
			return ScalarReal(-1);
		}

		RTRConfig::setConfigDb(*configDb);
		RTRRTRecordService* recordService = new RTRSSLRTRecordService(appId, serviceName, serverNameStr.c_str());
		if ( !recordService->active() ) {
			Rf_error("Error: %s\n" , recordService->text());
			exit(-1);
		}
		
		RTRTS1TimeSeriesService* tsService = new RTRTS1TimeSeriesService(*recordService, "IDN_TS1");
		tsService = new RTRTS1TimeSeriesService(*recordService, "IDN_TS1");
	    TSClientByCount* client;
		
		if (num > 0)
			client = new TSClientByCount(*tsService, symbol, num, period, debugOn);
		else 
			client = new TSClientByCount(*tsService, symbol, end, start, period, debugOn);
            
		RTRSelectNotifier::run();

		/*
		TODO this causes problems with returned records
		 - need to look at fixing thie leak
		delete client;
		delete tsService;
		delete recordService;
		delete configDb;*/

		if (client->dataReturned)
			return client->getFrame();
		else 
			return ScalarReal(0);
	}
}

/**
* Note: only a subset of these periodicity specifications are 
actually supported on IDN_TS1
*/
RTRTimeSeriesService::Period mnemonicToPeriod(const char* mnemonic) {
    RTRTimeSeriesService::Period p = RTRTimeSeriesService::Period::Weekly;
    if (strcmp(mnemonic, "d") == 0)
        p = RTRTimeSeriesService::Period::Daily;
    else if (strcmp(mnemonic, "c") == 0)
        p = RTRTimeSeriesService::Period::Calendar;
    else if (strcmp(mnemonic, "m") == 0)
        p = RTRTimeSeriesService::Period::Monthly;
    else if (strcmp(mnemonic, "q") == 0)
        p = RTRTimeSeriesService::Period::Quarterly;
    else if (strcmp(mnemonic, "y") == 0)
        p = RTRTimeSeriesService::Period::Yearly;
    else if (strcmp(mnemonic, "o") == 0)
        p = RTRTimeSeriesService::Period::DailyOpen;

    return p;
}
