/**
 * R/SFC time series interface. 
 * Based on the tsbycount example from the Reuters SFC SDK
 * Rory Winston <rory@theresearchkitchen.com>
 */

#include "rtr/rtrnotif.h"		
#include "tsbycount.h"

#include <sstream>

TSClientByCount::TSClientByCount(
								 RTRTimeSeriesService& service,
								 const char *symbol,
								 int numSamples,
								 RTRTimeSeriesService::Period period,
								 bool debugOn
								 ) : _series(service.timeSeries(symbol, period, *this)), 
								 _numSamples(numSamples), _byDate(false), _debugOn(debugOn), frame(NULL)
{
	((RTRTSValDefDb&) (service.tsValDefDb())).addClient(*this);
}

TSClientByCount::TSClientByCount(
								 RTRTimeSeriesService& service,
								 const char *symbol,
								 const char* startDate,
								 const char* endDate,
								 RTRTimeSeriesService::Period period,
								 bool debugOn
								 )
								 : _series(service.timeSeries(symbol, period, *this)), 
								 _byDate(true), _debugOn(debugOn), frame(NULL)
{
	((RTRTSValDefDb&) (service.tsValDefDb())).addClient(*this);
	_startDate = new RTRDateTime(startDate, RTRTRUE);
	_endDate= new RTRDateTime(endDate, RTRTRUE);
}


TSClientByCount::~TSClientByCount() {
	//_series = 0;
}

void TSClientByCount::processSeriesComplete(RTRTimeSeries& series) 
{ 
	int initialized = 0;
	
	if (!series.error() && series.count() > 0) {
		SEXP names;
		int cols = 0, rows = 0;
		int rowIndex = 0;

		// Preprocessing step : calculate the number of valid samples
		for (int i = series.lower(); i <= series.upper(); ++i) {
			RTRTimeSample& sample = series[i];
			if(sample.valid())
				rows++;	
		}

		if (_debugOn)
			Rprintf("Data contains %i rows (lower=%d, upper=%d)\n", rows, series.lower(), series.upper());

		for (int i = series.lower(); i <= series.upper(); ++i) 
		{
			// Initialize the data frame columns from a valid sample
			if (!initialized && series[i].valid()) 
			{
				RTRTimeSample& sample = series[i];
				cols = sample.upper() + 2; // extra column for 'Date' field
				
				if (_debugOn)
					Rprintf("Allocating data frame [%i rows, %i cols]\n", rows, cols);

				PROTECT(frame = allocVector(VECSXP, cols));
				PROTECT(names = allocVector(STRSXP, cols));

				SEXP dateColumn;
				PROTECT(dateColumn = allocVector(STRSXP, rows));
				SET_VECTOR_ELT(frame, 0, dateColumn);
				SET_STRING_ELT(names, 0, mkChar("Date"));

				for (int j = sample.lower(); j <= sample.upper(); ++j) 
				{
					int colIndex = j + 1;
					SEXP col;
					
					if (sample[j].dataType() < 2) 
						PROTECT(col = allocVector(REALSXP, rows));
					else
						PROTECT(col = allocVector(STRSXP, rows));

					SET_VECTOR_ELT(frame, colIndex, col);
					SET_STRING_ELT(names, colIndex, mkChar(sample[j].name())); 
				}

				initialized = 1;
			}
			
			RTRTimeSample& sample = series[i];

			if (sample.valid()) {
				stringstream ss;
				const RTRDateTime& dateTime = sample.dateTime();
				ss  << dateTime.monthNumber() << "/" << dateTime.dayOfMonth() << "/" << dateTime.year();
				SET_STRING_ELT(VECTOR_ELT(frame, 0), rowIndex, mkChar(ss.str().c_str()));
				for (int j = sample.lower(); j <= sample.upper(); ++j) 
				{
					int colIndex = j+1;
					RTRSeriesValue& value = sample[j];
					
					if (value.dataType() < 2) 
					{ 
						double* vec =  REAL(VECTOR_ELT(frame, colIndex));
						vec[rowIndex] = (double)value;
					}
					else 
					{
						SET_STRING_ELT(VECTOR_ELT(frame, colIndex), rowIndex, mkChar(value));
					}			
				}

				rowIndex++;
			}
		}

		setAttrib(frame, R_NamesSymbol, names);
		UNPROTECT(cols + 2);
	}
	else {
		if (_debugOn)
			Rprintf("Data had no rows\n");
		dataReturned = false;
		PROTECT(frame = ScalarInteger(-1));
		UNPROTECT(1);
	}

	RTREventNotifierInit::notifier->disable();
}

SEXP& TSClientByCount::getFrame() {
	if (frame == NULL) {
		PROTECT(frame = ScalarInteger(-1));
		UNPROTECT(1);
	}
	
	return frame; 
}

void TSClientByCount::processSeriesError(RTRTimeSeries& series)
{
	RTREventNotifierInit::notifier->disable();
	Rprintf("Error retrieving data: %s\n", series.text());
}

void TSClientByCount::processSeriesInfo(RTRTimeSeries& s)
{
}


void TSClientByCount::processDefDbUpdate(RTRTSValDefDb& db)
{ 
}


void TSClientByCount::processDefDbComplete(RTRTSValDefDb& db)
{

	if (_debugOn) {
		if (_byDate)
			Rprintf("Using date range [start=%i/%i/%i, end=%i/%i/%i]\n", 
			_endDate->dayOfMonth(),
			_endDate->monthNumber(),
			_endDate->year(),
			_startDate->dayOfMonth(),
			_startDate->monthNumber(),
			_startDate->year());
		else
			Rprintf("Requesting %i samples\n", _numSamples);
	}
	
	if (!_byDate)
		_series.setView(_numSamples);
	else {
		_series.setView(*_startDate, *_endDate);
	}

	if (!_series.error())
	{
		if (_series.complete())
			processSeriesComplete(_series);
		else
			processSeriesInfo(_series);
	}
	else
		processSeriesError(_series);
}
